package Autofont::Encoding;
use FileHandle;

my(%named_encodings);

sub new (;$$) {
    my $class = shift;
    my $self = { };
    bless $self, $class;
    if (@_) {
	$self->read(@_);
    } else {
	$self;
    }
}

sub named ($$) {
    my($class, $name) = @_;
    return $named_encodings{$name} if $named_encodings{$name};
    
    my($enc, $writable_dir);
    if ($name eq 'raw') {
	$enc = $class->new();
	$enc->{'raw'} = 1;
    } else {
	my($filename) = '';
	if (index($name, '.') <= index($name, '/')) {
	    $filename = `kpsewhich --format='PostScript header' $name.enc`;
	}
	if (!$filename) {
	    $filename = `kpsewhich --format='PostScript header' $name`;
	}
	chomp $filename;
	if (-r $filename) {
	    $enc = $class->new()->read($filename);
	}
    }

    if ($enc) {
	$enc->{'name'} = $name;
	$named_encodings{$name} = $enc;
    }

    # XXX
    if ($name =~ /^(.*)-sc$/) {
	my($base_name) = $1;
	my($base_enc) = $class->named($base_name);
	print STDERR "$writable_dir\n\n\n";
	if ($base_enc && $writable_dir
	    && (!$enc || -M $enc->{'filename'} > -M $base_enc->{'filename'})) {
	    $enc = $base_enc->derive('sc', $writable_dir);
	}
    }
    
    $enc;
}

sub calculate_map ($) {
    my($enc) = @_;
    my($map, $vec) = ({ }, $enc->{'vector'});
    for (my $i = 0; $i < @$vec; $i++) {
	$map->{$vec->[$i]} = $i;
    }
    $enc->{'map'} = $map;
}

sub read ($$) {
    my($enc, $fh) = @_;

    # read VPL file
    my($fh_provided) = (ref($fh));
    if (!$fh_provided) {
	my($fn) = $fh;
	$enc->{'filename'} = $fn;
	$fh = new FileHandle $fn, "r";
	return undef if !defined($fh);
    }
    local($/) = undef;
    my($v) = <$fh>;
    $fh->close if !$fh_provided;

    # canonicalize it
    # remove comments and then newlines
    $v =~ s/%.*$//mg;
    $v =~ tr/\n/ /;

    # get encoding name
    return undef if $v !~ m{\A\s*/([-.\w]+)\s*\[\s*(.*)\Z};
    $enc->{'psname'} = $1;
    $v = $2;

    # parse encoding
    my($vec) = [ ];
    while ($v =~ m{\A\s*/([-.\w]+)(.*)\Z}) {
	push @$vec, $1;
	$v = $2;
    }
    return undef if @$vec == 0;
    $enc->{'vector'} = $vec;

    $enc->calculate_map();
    $enc;
}

my %uc_accents = ( '' => 1, 'breve' => 1, 'caron' => 1, 'cedilla' => 1,
		   'circumflex' => 1, 'dotaccent' => 1, 'hungarumlaut' => 1,
		   'slash' => 1, 'ogonek' => 1, 'ring' => 1, 'tilde' => 1 );
my %uc_tranforms = ( 'ae' => 'AE', 'breve' => 'Breve', 'caron' => 'Caron',
		     'cedilla' => 'Cedilla', 'circumflex' => 'Circumflex',
		     'dotaccent' => 'Dotaccent', 'hungarumlaut' => 'Hungarumlaut', 'oe' => 'OE',
		     'ogonek' => 'Ogonek', 'ring' => 'Ring', 'tilde' => 'Tilde' );

sub derive ($$;$) {
    my($enc, $derivation, $dir) = @_;
    if ($derivation eq 'sc') {
	my($nenc) = $enc->new();
	$nenc->{'filename'} = $dir . "/" . $enc->{'filename'};
	die if $nenc->{'filename'} !~ /\.enc$/;
	$nenc->{'filename'} =~ s/\.enc$/-sc.enc/;
	
	$nenc->{'psname'} = $enc->{'psname'} . "-SC";

	my($vec) = [ ];
	foreach my $e (@{$enc->{'vector'}}) {
	    if ($e =~ /^[a-z](.*)$/ && $uc_accents{$1}) {
		push @$vec, ucfirst($e) . "small";
	    } elsif ($uc_transforms{$e}) {
		push @$vec, $uc_transforms{$e} . "small";
	    } else {
		push @$vec, $e;
	    }
	}
	$nenc->{'vector'} = $vec;
	
	$nenc->calculate_map();

	open OUT, ">" . $nenc->{'filename'} or die;
	print OUT "% Generated automatically by autofont from ", $enc->{'filename'}, "\n";
	print OUT "/", $nenc->{'psname'}, " [";
	for (my $i = 0; $i < 256; $i++) {
	    print OUT "\n " if ($i % 8) == 0;
	    print OUT " /", $vec->[$i];
	}
	print OUT "\n] def\n";
	close OUT;
	
	return $nenc;
    } else {
	return undef;
    }
}
	

sub raw ($) {
    my($enc) = @_;
    $enc->{'raw'};
}

sub psname ($) {
    my($enc) = @_;
    $enc->{'psname'};
}

sub filename ($) {
    my($enc) = @_;
    $enc->{'filename'};
}

sub map ($@) {
    my($enc) = shift @_;
    if (!wantarray && @_ == 1) {
	$enc->{'map'}->{ $_[0] };
    } else {
	my($m) = $enc->{'map'};
	map { $m->{$_} } @_;
    }
}

1;
